import gym
import numpy as np

from gym_multiagent.action import NoOp, POSSIBLE_ACTIONS
from gym_multiagent.world import World


ACTION_LOOKUP = dict(enumerate(POSSIBLE_ACTIONS))


class MultiAgentEnv(gym.Env):
    metadata = {"render.modes": ["human", "raw"]}

    def __init__(self, level=None):
        self.world = World(level)
        self.action_space = gym.spaces.Discrete(len(ACTION_LOOKUP))
        self.observation_space = gym.spaces.Tuple(
            [gym.spaces.Discrete(x) for x in self.world.level.shape + (self.world.numAgents + self.world.numBoxes,)]
        )
        self.goalsSolved = 0

    def step(self, actions):
        changes = []
        success = np.full(self.world.numAgents, True)

        for i, action in enumerate(map(ACTION_LOOKUP.get, actions)):
            if action == NoOp:
                continue

            agentLocation = self.world.agents[i]

            (moveLocation, boxLocation, newAgentLocation, newBoxLocation) = action(agentLocation)

            # Check box first to avoid index out of bounds error
            allowed = (boxLocation is None or self.world.hasBox(boxLocation)) and self.world.isFree(moveLocation)

            if allowed:
                changes.append((agentLocation, newAgentLocation))

                if boxLocation is not None:
                    changes.append((boxLocation, newBoxLocation))
            else:
                success[i] = False

        if changes:
            self.world.update(changes)

        goalsSolved = self.world.numGoalsSolved()

        observation = success
        reward = -0.1 + (self.goalsSolved - goalsSolved) * 10
        done = goalsSolved == self.world.numGoals

        self.goalsSolved = goalsSolved

        return observation, reward, done, {}

    def reset(self):
        self.world.reset()

    def render(self, mode="human"):
        if mode == "human":
            from gym_multiagent.render import render

            render(self.world)
        else:
            # Write current state to stdout as the default
            state_xy = np.where(self.world.state != " ")
            level = self.world.fixed.copy()
            level[state_xy] = self.world.state[state_xy]
            print("\n".join("".join(row) for row in level))

    def close(self):
        pass



class Point(tuple):
    def __new__(cls, x, y):
        return tuple.__new__(cls, (x, y))

    @property
    def x(self):
        return self.__getitem__(0)

    @property
    def y(self):
        return self.__getitem__(1)

    def __add__(self, other):
        return Point(self.x + other.x, self.y + other.y)

    def __sub__(self, other):
        return Point(self.x - other.x, self.y - other.y)

    def __repr__(self):
        return f"Point{tuple.__repr__(self)}"



from gym_multiagent.point import Point


class Action(tuple):
    """Action base class
    
    Each action is associated with a type, a move direction and optionally a box direction.
    """

    def __new__(cls, moveDirection, boxDirection=None):
        assert moveDirection in LOCATION_CHANGE
        assert boxDirection is None or boxDirection in LOCATION_CHANGE
        return tuple.__new__(cls, (moveDirection, boxDirection))

    @property
    def actionType(self):
        return self.__class__.__name__

    @property
    def moveDirection(self):
        return self.__getitem__(0)

    @property
    def boxDirection(self):
        return self.__getitem__(1)

    def __repr__(self):
        if self.boxDirection is None:
            return f"{self.actionType}({self.moveDirection})"
        return f"{self.actionType}({self.moveDirection},{self.boxDirection})"

    def __call__(self, agentLocation):
        """Returns a tuple (moveLocation, boxLocation, newAgentLocation, newBoxLocation)
        - `moveLocation` indicates which location should be free to perform this action.
        - `boxLocation` indicates which location there should be a box to perform this action.
        - `newAgentLocation` indicates where the agent will move after performing this action.
        - `newBoxLocation` indicates where the box will move after performing this action.
        """
        raise NotImplementedError()


class Move(Action):
    def __new__(cls, moveDirection):
        return Action.__new__(cls, moveDirection)

    def __call__(self, agentLocation):
        """Return a Move tuple (see super class)
        - The agent moves according to `moveDirection`
        """
        newAgentLocation = agentLocation + LOCATION_CHANGE[self.moveDirection]
        moveLocation = newAgentLocation
        return (moveLocation, None, newAgentLocation, None)


class Push(Action):
    def __new__(cls, moveDirection, boxDirection):
        return Action.__new__(cls, moveDirection, boxDirection)

    def __call__(self, agentLocation):
        """Return a Push tuple (see super class)
        - The box is in the `boxDirection` relative to the agent
        - The agent moves to the current box location
        - The box moves according to `moveDirection`
        """
        boxLocation = agentLocation + LOCATION_CHANGE[self.boxDirection]
        newAgentLocation = boxLocation
        newBoxLocation = boxLocation + LOCATION_CHANGE[self.moveDirection]
        moveLocation = newBoxLocation
        return (moveLocation, boxLocation, newAgentLocation, newBoxLocation)


class Pull(Action):
    def __new__(cls, moveDirection, boxDirection):
        return Action.__new__(cls, moveDirection, boxDirection)

    def __call__(self, agentLocation):
        """Return a Pull tuple (see super class)
        - The box is in the `boxDirection` relative to the agent
        - The agent moves according to `moveDirection`
        - The box moves to the current agent location
        """
        boxLocation = agentLocation + LOCATION_CHANGE[self.boxDirection]
        newAgentLocation = agentLocation + LOCATION_CHANGE[self.moveDirection]
        newBoxLocation = agentLocation
        moveLocation = newAgentLocation
        return (moveLocation, boxLocation, newAgentLocation, newBoxLocation)


# fmt: off
N = 'N'; S = 'S'; W = 'W'; E = 'E'; NoOp = 'NoOp'

LOCATION_CHANGE = {
    N: Point(-1,  0),
    S: Point( 1,  0),
    W: Point( 0, -1),
    E: Point( 0,  1),
}

POSSIBLE_ACTIONS = [
    NoOp,
    Move(N)  , Move(S)  , Move(W)  , Move(E)  ,
    Push(N,N), Push(S,S), Push(W,N), Push(E,S),
    Push(N,W), Push(S,W), Push(W,S), Push(E,W),
    Push(N,E), Push(S,E), Push(W,W), Push(E,E),
    Pull(N,S), Pull(S,N), Pull(W,N), Pull(E,N),
    Pull(N,W), Pull(S,W), Pull(W,S), Pull(E,S),
    Pull(N,E), Pull(S,E), Pull(W,E), Pull(E,W),
]
# fmt: on



from typing import Iterable, Tuple

import numpy as np

from gym_multiagent.point import Point

DEFAULT_LEVEL = """
++++++++
+0+  cb+
+ A+ ad+
+ +B+  +
+  DC+ +
+ ++++ +
+      +
++++++++
"""


class World(object):
    def __init__(self, level: str = None):
        if level is None:
            level = DEFAULT_LEVEL

        self.level = np.array([[c for c in line] for line in level.split("\n") if line])
        self.state = np.full_like(self.level, " ")
        self.fixed = np.full_like(self.level, " ")

        agents_xy = np.where(np.char.isdigit(self.level))
        boxes_xy = np.where(np.char.isupper(self.level))
        goals_xy = np.where(np.char.islower(self.level))
        walls_xy = np.where(self.level == "+")

        self.state[agents_xy] = self.level[agents_xy]
        self.state[boxes_xy] = self.level[boxes_xy]

        self.fixed[goals_xy] = self.level[goals_xy]
        self.fixed[walls_xy] = "+"

        self.agents = {int(c): Point(x, y) for c, x, y in zip(self.level[agents_xy], *agents_xy)}
        self.goals = {Point(x, y): c.upper() for c, x, y in zip(self.level[goals_xy], *goals_xy)}

        self.numAgents = len(agents_xy[0])
        self.numBoxes = len(boxes_xy[0])
        self.numGoals = len(goals_xy[0])

        self.initialAgents = self.agents.copy()
        self.initialState = self.state.copy()

    def reset(self):
        self.agents = self.initialAgents.copy()
        self.state = self.initialState.copy()

    def update(self, changes: Iterable[Tuple[Point, Point]]):
        fromLocations, toLocations = zip(*changes)
        freedLocations = set(fromLocations) - set(toLocations)

        from_xy = list(zip(*fromLocations))
        to_xy = list(zip(*toLocations))
        freed_xy = list(zip(*freedLocations))

        agents_idx = np.where(np.char.isdigit(self.state[from_xy]))[0]

        for agent_idx in agents_idx:
            agentLocation = fromLocations[agent_idx]
            newAgentLocation = toLocations[agent_idx]
            self.agents[int(self.state[agentLocation])] = newAgentLocation

        self.state[to_xy] = self.state[from_xy]
        self.state[freed_xy] = " "

    def numGoalsSolved(self):
        return sum(self.state[location] == c for location, c in self.goals.items())

    def isFree(self, location: Point):
        return self.fixed[location] != "+" and self.state[location] == " "

    def hasBox(self, location: Point):
        return self.state[location].isalpha()


if __name__ == "__main__":
    world = World()



import sys

import numpy as np

from gym_multiagent.world import World

try:
    import matplotlib.pyplot as plt
except:
    print("matplotlib is required to render in 'human' mode")
    print("change to 'raw' or run 'pip install matplotlib'")
    sys.exit(-1)


def Rectangle(x, y, size=1, **kwargs):
    return plt.Rectangle([x - size / 2, y - size / 2], size, size, **kwargs)


def Circle(x, y, size=1, **kwargs):
    return plt.Circle([x, y], size / 2, **kwargs)


def render(world: World):
    ax = plt.gca()

    for (x, y) in np.ndindex(world.level.shape):
        f = world.fixed[(x, y)]
        s = world.state[(x, y)]

        if f == "+":
            ax.add_patch(Rectangle(y, x, fc="black", ec="black"))
        elif f.isalpha():
            ax.add_patch(Rectangle(y, x, fc="lime" if f.upper() == s else "yellow", ec="black"))
            ax.text(y, x, f, fontsize=16, fontweight="bold", ha="center", va="center")
        else:
            ax.add_patch(Rectangle(y, x, fc="lightgray", ec="black"))

        if s.isalpha():
            ax.add_patch(Rectangle(y, x, size=0.75, fc="blue", ec="blue"))
            ax.text(y, x, s, color="white", fontsize=16, fontweight="bold", ha="center", va="center")
        elif s.isnumeric():
            ax.add_patch(Circle(y, x, size=0.75, fc="blue", ec="blue"))
            ax.text(y, x, s, color="white", fontsize=16, fontweight="bold", ha="center", va="center")

    ax.margins(0)
    ax.set_aspect("equal", "box")
    ax.invert_yaxis()
    ax.set_axis_off()
    plt.tight_layout()
    plt.show()


if __name__ == "__main__":
    world = World()
    render(world)



from setuptools import setup, find_packages

setup(
    name="gym_multiagent",
    version="1.0.0",
    license="MIT",
    description="Sokoban-inspired multi-agent environment for OpenAI Gym",
    author="helgehatt",
    url="https://github.com/helgehatt/gym-multiagent",
    keywords=["multiagent", "openai", "gym", "environment"],
    install_requires=["gym", "numpy"],
    packages=find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)
